﻿namespace HeuristicLab.Problems.ProgramSynthesis.Push.Expressions {
  using System;
  using System.Collections.Generic;

  using HeuristicLab.Problems.ProgramSynthesis.Push.Attributes;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Interpreter;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Stack;

  /// <summary>
  ///     Defines the name on top of the NAME stack as an instruction that will push the top item of the T stack onto the
  ///     EXEC stack.
  /// </summary>
  /// <typeparam name="T">Stacktype</typeparam>
  public abstract class DefineExpression<T> : StatelessExpression {
    protected void Eval(
      IStack<T> stack,
      IStack<string> nameStack,
      IDictionary<string, Expression> customExpressions,
      Func<T, Expression> creator) {
      // not enough arguments on stack
      if ((nameStack.Count == 0) || (stack.Count < 1)) return;

      var name = nameStack.Pop();
      var expression = creator(stack.Top);

      if (customExpressions.ContainsKey(name)) customExpressions[name] = expression;
      else customExpressions.Add(name, expression);
    }
  }

  [PushExpression(StackType.Code, "CODE.DEFINE")]
  public class CodeDefineExpression : DefineExpression<Expression> {
    public override void Eval(IPushGpInterpreter interpreter) {
      if (interpreter.Configuration.TopLevelPushCode && (interpreter.CodeStack.Count < 2)) return;

      this.Eval(
        interpreter.CodeStack,
        interpreter.NameStack,
        interpreter.CustomExpressions,
        v => new ExecPushExpression(v));

      interpreter.CodeStack.Pop();
    }
  }

  [PushExpression(StackType.Exec, "EXEC.DEFINE")]
  public class ExecDefineExpression : DefineExpression<Expression> {
    public override void Eval(IPushGpInterpreter interpreter) {
      if (interpreter.ExecStack.Count < 2) return;

      this.Eval(
        interpreter.ExecStack,
        interpreter.NameStack,
        interpreter.CustomExpressions,
        v => new ExecPushExpression(v));

      interpreter.ExecStack.Pop();
    }
  }

  [PushExpression(StackType.Float, "FLOAT.DEFINE")]
  public class FloatDefineExpression : DefineExpression<double> {

    public override void Eval(IPushGpInterpreter interpreter) {
      this.Eval(
        interpreter.FloatStack,
        interpreter.NameStack,
        interpreter.CustomExpressions,
        v => new FloatPushExpression(v));
    }
  }

  [PushExpression(StackType.Integer, "INTEGER.DEFINE")]
  public class IntegerDefineExpression : DefineExpression<long> {
    public override void Eval(IPushGpInterpreter interpreter) {
      this.Eval(
        interpreter.IntegerStack,
        interpreter.NameStack,
        interpreter.CustomExpressions,
        v => new IntegerPushExpression(v));
    }
  }

  [PushExpression(StackType.Boolean, "BOOLEAN.DEFINE")]
  public class BooleanDefineExpression : DefineExpression<bool> {
    public override void Eval(IPushGpInterpreter interpreter) {
      this.Eval(
        interpreter.BooleanStack,
        interpreter.NameStack,
        interpreter.CustomExpressions,
        v => new BooleanPushExpression(v));
    }
  }
}