﻿namespace HeuristicLab.Problems.ProgramSynthesis.Push.Expressions {
  using System;

  using HeuristicLab.Problems.ProgramSynthesis.Push.Attributes;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Interpreter;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Stack;

  /// <summary>
  ///     Pushes the sum of the top two items.
  /// </summary>
  [PushExpression(StackType.Float, "FLOAT.+")]
  public class FloatAddExpression : PushResultExpression<double> {
    public override void Eval(IPushGpInterpreter interpreter) {
      Eval(interpreter.FloatStack, 2, values => values[0] + values[1]);
    }
  }

  /// <summary>
  ///     Pushes the difference of the top two items; that is, the second item minus the top item.
  /// </summary>
  [PushExpression(StackType.Float, "FLOAT.-")]
  public class FloatSubtractExpression : PushResultExpression<double> {
    public override void Eval(IPushGpInterpreter interpreter) {
      this.Eval(interpreter.FloatStack, 2, values => values[0] - values[1]);
    }
  }

  /// <summary>
  ///     Pushes the product of the top two items.
  /// </summary>
  [PushExpression(StackType.Float, "FLOAT.*")]
  public class FloatMultiplyExpression : PushResultExpression<double> {
    public override void Eval(IPushGpInterpreter interpreter) {
      this.Eval(interpreter.FloatStack, 2, values => values[0] * values[1]);
    }
  }

  /// <summary>
  ///     Pushes the quotient of the top two items; that is, the second item divided by the top item.
  ///     If the top item is zero this acts as a NOOP.
  /// </summary>
  [PushExpression(StackType.Float, "FLOAT./")]
  public class FloatDivideExpression : PushResultExpression<double> {
    public override void Eval(IPushGpInterpreter interpreter) {
      this.Eval(interpreter.FloatStack, 2, values => values[0] / values[1], 0);
    }
  }

  /// <summary>
  ///     Pushes the second stack item modulo the top stack item. If the top item is zero this acts as a NOOP.
  ///     The modulus is computed as the remainder of the quotient, where the quotient has first been truncated toward
  ///     negative infinity.
  ///     (This is taken from the definition for the generic MOD function in Common Lisp, which is described
  ///     for example at http://www.lispworks.com/reference/HyperSpec/Body/f_mod_r.htm.)
  /// </summary>
  [PushExpression(StackType.Float, "FLOAT.%")]
  public class FloatModuloExpression : PushResultExpression<double> {
    public override void Eval(IPushGpInterpreter interpreter) {
      this.Eval(interpreter.FloatStack, 2, values => values[0] % values[1], 0);
    }
  }

  /// <summary>
  ///     Pushes the minimum of the top two items.
  /// </summary>
  [PushExpression(StackType.Float, "FLOAT.MIN")]
  public class FloatMinExpression : PushResultExpression<double> {
    public override void Eval(IPushGpInterpreter interpreter) {
      this.Eval(interpreter.FloatStack, 2, values => Math.Min(values[0], values[1]));
    }
  }

  /// <summary>
  ///     Pushes the maximum of the top two items.
  /// </summary>
  [PushExpression(StackType.Float, "FLOAT.MAX")]
  public class FloatMaxExpression : PushResultExpression<double> {
    public override void Eval(IPushGpInterpreter interpreter) {
      this.Eval(interpreter.FloatStack, 2, values => Math.Max(values[0], values[1]));
    }
  }

  /// <summary>
  ///     Pushes TRUE onto the BOOLEAN stack if the second item is less than the top item, or FALSE otherwise.
  /// </summary>
  [PushExpression(StackType.Float, "FLOAT.<")]
  public class FloatSmallerThanExpression : PushResultExpression<double> {
    public override void Eval(IPushGpInterpreter interpreter) {
      this.Eval(interpreter.FloatStack, interpreter.BooleanStack, 2, values => values[0] < values[1]);
    }
  }

  /// <summary>
  ///     Pushes TRUE onto the BOOLEAN stack if the second item is greater than the top item, or FALSE otherwise.
  /// </summary>
  [PushExpression(StackType.Float, "FLOAT.>")]
  public class FloatGreaterThanExpression : PushResultExpression<double> {
    public override void Eval(IPushGpInterpreter interpreter) {
      this.Eval(interpreter.FloatStack, interpreter.BooleanStack, 2, values => values[0] > values[1]);
    }
  }

  /// <summary>
  ///     Pushes the sine of the top item.
  /// </summary>
  [PushExpression(StackType.Float, "FLOAT.SIN")]
  public class FloatSineExpression : PushResultExpression<double> {
    public override void Eval(IPushGpInterpreter interpreter) {
      this.Eval(interpreter.FloatStack, 1, values => Math.Sin(values[0]));
    }
  }

  /// <summary>
  ///     Pushes the cosine of the top item.
  /// </summary>
  [PushExpression(StackType.Float, "FLOAT.COS")]
  public class FloatCosineExpression : PushResultExpression<double> {
    public override void Eval(IPushGpInterpreter interpreter) {
      this.Eval(interpreter.FloatStack, 1, values => Math.Cos(values[0]));
    }
  }

  /// <summary>
  ///     Pushes 1 if the top BOOLEAN is TRUE, or 0 if the top BOOLEAN is FALSE.
  /// </summary>
  [PushExpression(StackType.Float, "FLOAT.FROMBOOLEAN")]
  public class FloatFromBooleanExpression : StatelessExpression {
    public override void Eval(IPushGpInterpreter interpreter) {
      if (interpreter.BooleanStack.Count == 0) return;

      var condition = interpreter.BooleanStack.Pop();
      var value = condition ? 1 : 0;

      interpreter.FloatStack.Push(value);
    }
  }

  /// <summary>
  ///     Pushes a floating point version of the top INTEGER.
  /// </summary>
  [PushExpression(StackType.Float, "FLOAT.FROMINTEGER")]
  public class FloatFromIntegerExpression : StatelessExpression {
    public override void Eval(IPushGpInterpreter interpreter) {
      if (interpreter.IntegerStack.Count == 0) return;

      var value = (double)interpreter.IntegerStack.Pop();

      interpreter.FloatStack.Push(value);
    }
  }
}